//
//	positionLocator.cc: test the PositionLoator class
//
//	DESCRIPTION:
//		Test the PositionLocator class with the predicted orbit file
//		in ../odf. Data in the file is valid from 2000-01-21T01:39:56
//		till 2002-09-10T00:00:00. The S/C position and velocity is
//		checked for 4 different times in the interval
//			2000-01-21T12:32:00 - 2000-01-27T08:47:00
//		It has been check that the results are identical to the output
//		of the Orbit Access S/W routine (orbita) from FD.
//

#include "OalServer.h"

#include <iostream>
#include <cassert>
#include <cmath>
#include "common.h"
#include "errstr.h"
#include "STime.h"

static bool isClose(double a, double b)
{
    return fabs(fabs(a)/fabs(b)-1.) < 1.e-7;
}

int main()
{
	const Odf *odf = oalServer.newOdf("../odf");
	errstr<<*odf<<Sas::message(Msg::App, Msg::Verbose);

	const OdfSet *orbitSet =
				oalServer.odf()->selectSet(Xmm::SC, Xmm::PREDICTED_ORBIT);

	STime startTime("2000-01-21T12:32:00"),
		  endTime("2000-01-27T08:47:00");
	const XmmPositionLocator *posloc = new XmmPositionLocator(orbitSet);

	//
	//	check position/velocity at a couple of times
	//
	unsigned Ntimes = 3;
	double trueStateVecs[][6] = {
		{ -16020.45, -6472.1011, -12310.45, 3.9136032, -1.9162562, 3.513118},
		{ 10627.65, 8003.7472, 7526.9643, 0.87417344, 6.721378, -0.28988464},
		{ 12691.524, 3891.4821, 9788.6817, 1.6182822, 6.2926732, 0.36422967},
		{ 352.48876, -14334.617, 2491.6739, 5.1857164, 2.7114092, 3.8188753},
		{ -17659.655, -15510.587, -12007.664, 3.6440079,-1.0619844,3.1373652}};

	double deltaT = (endTime-startTime)/DBL(Ntimes);
	for (unsigned n=0; n<=Ntimes+1; ++n) {
		STime t(startTime + DBL(n)*deltaT);
		cout<<"State vector at t="<<t<<" : ";
		try {
			StateVector stateVector = posloc->stateVector(t);
			cout.precision(8);
			cout<<stateVector<<endl;
			for (unsigned i=0; i<6; ++i)
				assert(isClose(stateVector[i], trueStateVecs[n][i]));
		} catch(...) {
			cerr<<"... not available\n";
			assert(1==2);	//	should never happen
		}
	}

	try {
		StateVector stateVector = posloc->stateVector(1.e10);
		assert(1==2);	//	should not happen
	} catch (...) {
	}

	cout<<"\nAll tests successfully passed.\n";
	return 0;

}
