//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OdfInfo.h	--- Observation Data File INFormation
//
//	DESCRIPTION:
//		Various classes are delcared which provide information about the
//		ODF being pointed at via the environment variable SAS_ODF. The
//		queriable information include:
//
//			o global ODF attributes (rev no, location, start, stop times etc.)
//			o list of science exposures
//			o list of data sets associated with given exposure
//			o proposal info (source name + coords)
//
//	CHANGE HISTORY:
//		UL, 2001-07-02: 1.0.0: - initial version
//
//	AUTHOR:
//		U. Lammers, Mon Jul  2 12:00:02 MET DST 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/oal/odfbrowse/OdfInfo.h,v 1.4 2014/04/15 15:54:31 aibarra Exp $
///

#ifndef ODF_INFO_H
#define ODF_INFO_H

#include <string>
#include <vector>

#include "Xmm.h"
#include "STime.h"

namespace OdfInfo {

//
//	---( Server )-------------------------------------------------------------
//
class Odf;
class Server {
public:
	virtual const Odf *odf() const = 0;
		//
		//	PURPOSE:
		//		Obtain handle to ODF pointed at via SAS_ODF
		//


	virtual const Odf *odf(const std::string &location) const = 0;
		//
		//	PURPOSE:
		//		Obtain handle to ODF located in directory <location> which
		//		must contain a summary file "*SUM.SAS" created by
		//		odfingest
		//


	virtual ~Server() {}
};

extern Server *server;	//	single global OdfInfo::Server instance


//
//	---( Odf )-----------------------------------------------------------------
//
class TimedEvent {
public:
	virtual STime	start()	const = 0;
	virtual STime	stop()	const = 0;
		//
		//	PURPOSE:
		//		Obtain start/stop time of timed event
		//

	virtual	~TimedEvent() {}
};


class ProposalData;
class ObservationExposureList;
class Odf : public TimedEvent {
public:
	virtual std::string		location()		const = 0;	//	Where is it?
	virtual unsigned short	revolution()	const = 0;	//	What is the rev no?
	virtual std::string		proposalId()	const = 0;	//	6-char proposal ID
	virtual std::string		observationId()	const = 0;	//	4-char obs ID
	virtual bool			isSlew()		const = 0;	//	Is this a slew OD?

	virtual const ProposalData
							&proposalData()	const = 0;
												//	get handle to proposal data

	virtual const ObservationExposureList
						&observationExposures(Xmm::Instrument ins) const = 0;
												//	get handle to time-ordered
												//	list of exposures
												//	of instrument <ins>
};


//
//	---( ProposalData )--------------------------------------------------------
//
class ProposalData {
public:
	virtual std::string	GoName()		const = 0;
	virtual std::string	GoInstitute()	const = 0;
	virtual std::string	GoAddress()		const = 0;
	virtual std::string	GoCountry()		const = 0;
	virtual std::string	GoEmail()		const = 0;
	virtual std::string	AO()			const = 0;	//	Announcement of
													//	Oportunity, e.g. "AO1"
	virtual std::string	scienceType()	const = 0;

	virtual std::string	targetName()	const = 0;
	virtual double	targetRa()			const = 0;	//	[deg]
	virtual double	targetDec()			const = 0;	//	[deg]
	virtual double	obsDuration()		const = 0;	//	[s]
	virtual double	boresightRa()		const = 0;	//	[deg]
	virtual double	boresightDec()		const = 0;	//	[deg]

	//	lower/upper/ref of position angle
	virtual std::string	posAngleConstLo() const = 0;
	virtual std::string	posAngleConstHi() const = 0;
	virtual std::string	posAngleOrigRef() const = 0;

	//	instrument priorities
	virtual short	emos1Prio()		const = 0;
	virtual short	emos2Prio()		const = 0;
	virtual short	epnPrio()		const = 0;
	virtual short	rgs1Prio()		const = 0;
	virtual short	rgs2Prio()		const = 0;
	virtual short	omPrio()		const = 0;

	virtual	~ProposalData() {}
};


//
//	---( Exposure/List )-------------------------------------------------------
//
class OdfDataSetList;
class ObservationExposure : public TimedEvent {
public:
	enum Type { SCIENCE, CALIBRATION };
	virtual Type			type()		const = 0;

	virtual unsigned		counter()	const = 0;	//	exposure counter
	virtual bool			scheduled()	const = 0;	//	Is it a scheduled one?

	virtual Xmm::Instrument	instrument()const = 0;
	virtual DataMode		dataMode()	const = 0;	//	data mode
	virtual Mode			mode()		const = 0;	//	instrument mode
	virtual Filter			filter()	const = 0;	//	filter

	virtual std::string		id()		const = 0;
											//	return unique ID of this
											//	exposure of the form
											//		IIFnnn
											//	II	: instrument id
											//	F	: scheduled flag ('S'/'U')
											//	nnn	: sequence number (>=1)

	virtual const OdfDataSetList &associatedScienceSets() const = 0;
	virtual const OdfDataSetList &associatedAuxiliarySets() const = 0;
											//	obtain handle to list of
											//	associated list of science/
											//	auxiliary (engineering/HK) sets
};

class ObservationExposureList :
						public virtual std::vector<ObservationExposure *> {};


//
//	---( OdfDataSet/List )-----------------------------------------------------
//
class OdfDataSet {
public:
	//gcc 4.8.2 porting. Virtual destructor added
	virtual ~OdfDataSet() {}
	virtual std::string		name()			const = 0;
	virtual std::string		dirname()		const = 0;
	virtual std::string		basename()		const = 0;
	virtual unsigned		revolution()	const = 0;
	virtual std::string		proposalId()	const = 0;
	virtual Xmm::Instrument	instrument()	const = 0;
	virtual bool			scheduled()		const = 0;
	virtual unsigned		exposure()		const = 0;
	virtual unsigned short	ccd()			const = 0;
	virtual unsigned		node()			const = 0;
	virtual Xmm::DataMode	dataMode()		const = 0;

	enum Type { TNONE = -1, EVENT_LIST, IMAGE, AUXILIARY, HOUSKEEPING,
				SPACECRAFT, SUMMARY };
	virtual Type			type()			const = 0;

	enum Format { FNONE = -1, COMPRESSED_FITS, COMPRESSED_ASCII, FITS, ASCII };
	virtual Format			format()		const = 0;

	virtual bool			isValid()		const = 0;
};

class OdfDataSetList : public std::list<OdfDataSet *> {};

}

#endif
