//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OalServer.cc	--- implementation of OalServer classes
//
//	DESCRIPTION:
//		Implementation part of OalServer
//
//	CHANGE HISTORY:
//		UL, 1998-07-24: 1.0.0: - first version extracted from old OalAtoms.cc
//		UL, 1999-08-20: 1.1.0: - added diagnostic messages
//		UL, 1999-10-29: 1.2.0: - reset warp-around counter of
//								 InstrFrameCounter<> objects if ccd/node id
//								 has changed in state
//		UL, 1999-11-02: 1.3.0: - bail out in setCcdNr()/setNodeNr() on
//								 seeing invalid numbers
//		UL, 1999-12-17: 1.4.0: - implemented oalServer<<block/set
//								 functions
//		UL, 2000-04-13: 1.5.0: - made data mode list ODF ICD 2.4 compliant
//		UL, 2000-05-19: 1.6.0: - more diagnostic messages from setState()
//							   - prepending ODF directory name to data set
//								 name only of not found in current directory
//		UL, 2000-06-07: 1.7.0: - instantiate TimeCorrelator with much
//								 relaxed RMS criterion for OBT-MET fit
//		UL, 2000-11-09: 1.8.0: - several changes to fix SPR 1825: state
//								 variable dataMode is no longer obtained
//								 from contents of DATATYPE keyword but
//								 from asking the ODF for the data mode
//								 of the exposure with a given number
//		UL, 2000-12-07: 1.9.0: - OBT-MET fit parameters can be given in
//								 environment variable
//									SAS_OBT_MET_FIT
//		UL, 2000-12-08: 1.9.1: - a few trival changes to make the optimizer
//								 silent
//		UL, 2000-12-12: 1.9.2: - relaxed RMS criterion of OBT-MET fit to 1.-2
//		UL, 2000-08-10: 2.0.0: - various changes related to addition of
//								 exposureId in OAL state
//		UL, 2001-10-22: 2.1.0: - scheduled flag introduced into state; gets
//								 now set as part of setState(); this required
//								 extraction of scheduled flag from data set
//								 attribute 'FILENAME'
//		UL, 2001-12-05: 2.2.0: - added OalStateable::instrExposureId() - can
//								 also be given ODF data set name
//		UL, 2002-02-26: 2.3.0: - using new Odf::findExposure() to fix problem
//								 in SSC-SPR-2777
//		UL, 2002-03-08: 2.4.0: - make operator<<(OalServer,
//												 vector<Attributable>)
//								 robust against missing FILENAME attribute
//								 in dataset
//		UL, 2002-03-11: 2.4.1: - fixed problem introduced in 2.4.0
//		UL, 2002-03-22: 2.5.0: - changed defaults of SAS_OBT_MET_FIT from
//									0,1,1.e-2,.9 to 0,4,5.e-5,.9
//							   - added igngshov field to SAS_OBT_MET_FIT
//		JBa,2002-07-02: 2.5.1: - DEVELOPMENT-TRACK! Usage of sstream
//		UL, 2003-02-07: 2.5.2: - various changes related to new
//								 piecewise-linear TC data interpolation
//		UL, 2003-05-02: 2.6.0: - small change to fix problem with S/U
//								 exposure confusion (seen e.g. with
//								 ODF 0591_0144900101)
//		UL, 2003-06-18: 2.7.0: - added new environment variable SAS_ATTITUDE
//								 (see doc for details)
//
//	AUTHOR:
//		U. Lammers, Fri Jul 24 09:52:44 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/oal/src/OalServer.cc,v 1.57 2013/01/18 13:57:28 rsaxton Exp $
//
#include <sstream>
#include <fstream>

#include "ExceptionErrorHandler.h"
#include "OalServer.h"
#include "DalUtils.h"
#include "Tracer.h"
#include "errstr.h"
#include "stringutil.h"
#include "OdfDesignators.h"

using namespace std;

OalServer &OalServer::instance()
{
	static OalServer *oalServerInstance = new OalServer;

	return *oalServerInstance;
} // oalServerFunc()


//
//	---( OalStateable )----------------------------------------
//
void OalStateable::instrument(Xmm::Instrument instrumentId) const
{
	_instrument = instrumentId;
} // OalStateable::instrument()


void OalStateable::dataMode(Xmm::DataMode dataMode) const
{
	_dataMode = dataMode;
} // OalStateable::dataMode()


void OalStateable::exposureIndex(unsigned exposureIndex) const
{
	if (exposureIndex >= 1)
		_exposureIndex = exposureIndex;
	else
		errstr<<"Exposure index must be greater than 0"
			  <<Sas::error("IllegalExposureNumber");

	//	need to reset wraps around counter / reset time
	//	in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::reset();
	InstrFrameCounterVector<RGScam>::reset();
	InstrFrameCounterVector<OMcam>::reset();
	/* AI SOC-SCR-284
	 * ERM time conversion implementation.
	 */
	InstrFrameCounterVector<ERMcam>::reset();

	//	define exposure number
	const Exposure *exp =
		OalServer::instance().odf()->exposure(_instrument, _dataMode,
											  _exposureIndex, Odf::EXP_OAL,
											  true, _scheduled);
	_exposureNumber = exp->counter();
	_scheduled = exp->scheduled();
} // OalStateable::exposureIndex()


//
//	exposureId is a four-character string 'Xnnn' with X being the
//	scheduled flag ('S' or 'U') and nnn being the ODF exposure designator
//	e.g. exposureId = "S004" or "U571"; the latter happens to be also
//	the name of a good movie.
//
void OalStateable::exposureId(const string &expId) const
{
	string id(expId);
	if (id.length() > 4) {
		//	assume this is an ODF set name
		InstrumentExposureIdentifier insdes(id);
		id = insdes.exposureId();
	}

	if (id.length() != 4)
		errstr<<"Given exposure designator `"<<id<<"' does not match "
				"expected pattern `S|U[0-9][0-9][0-9]'"
			  <<Sas::error("InvalidExposureDesignator");

	_scheduled = toupper(id[0]) == 'S';
	_exposureNumber = toNumber<unsigned>(id.substr(1, 3));
	Xmm::DataMode mode = oalServer.odf()->whatMode(_instrument,
												   _exposureNumber,
												   _scheduled);
	oalServer.dataMode(mode);
	const Exposure *exp = oalServer.odf()->exposure(_instrument, mode,
													_exposureNumber,
													Odf::EXP_ODF, true,
													_scheduled);
	oalServer.exposureIndex(exp->index()+1);
} // OalStateable::exposureId()


//
//	instrExposureId is a six-character string 'IIXnnn' with II being
//	the instrument designator and 'Xnnn' the exposureId (see above);
//	can also be given a full ODF set name in which case the relevant
//	bits are extracted
//
void OalStateable::instrExposureId(const string &insExpId) const
{
	InstrumentExposureIdentifier insdes(insExpId);
	instrument(insdes.instrument());
	exposureId(insdes.exposureId());
} // OalStateable::instrExposureId()


string OalStateable::instrExposureId() const
{
	stringstream s;
	s<<Xmm::toString(instrument(), true)<<(scheduled() ? 'S' : 'U')
	 <<setw(3)<<setfill('0')<<exposureNumber();

	return s.str();
} // OalServer::instrExposureId()


void OalStateable::scheduled(bool scheduled) const
{
	_scheduled = scheduled;
} // OalStateable::scheduled()


void OalStateable::ccd(Ccd ccd) const
{
	if (ccd< 1 || ccd>XmmPayLoad::instrument(_instrument).nCcds())
		errstr<<"Cannot set CCD number to "<<ccd<<" - valid range is [1, "
			  <<XmmPayLoad::instrument(_instrument).nCcds()<<"]"
			  <<Sas::error("NoSuchCcd");

	_ccd = ccd;

	//	need to reset wraps around counter in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::resetWraps();
	InstrFrameCounterVector<RGScam>::resetWraps();
	InstrFrameCounterVector<OMcam>::resetWraps();
	/* AI SOC-SCR-284
	 * ERM time conversion implementation.
	 */
	InstrFrameCounterVector<ERMcam>::resetWraps();
} // OalStateable::ccd()


void OalStateable::node(Node node) const
{
	if (UNT(node)>XmmPayLoad::instrument(_instrument).nNodes())
		errstr<<"Cannot set node ID to "<<node<<" - valid range is [0, "
			  <<XmmPayLoad::instrument(_instrument).nNodes()-1<<"]"
			  <<Sas::error("NoSuchNode");

	_node = node;

	//	need to reset wraps around counter in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::resetWraps();
	InstrFrameCounterVector<RGScam>::resetWraps();
	InstrFrameCounterVector<OMcam>::resetWraps();
	/* AI SOC-SCR-284
	 * ERM time conversion implementation.
	 */
	InstrFrameCounterVector<ERMcam>::resetWraps();
} // OalStateable::node()


void OalStateable::attitudeFromStr(bool attitudeFromStr) const
{
	_attitudeFromStr = attitudeFromStr;
} // OalStateable::setAttitudeFromStr()

void OalStateable::timeJumpThreshold(unsigned timeJumpThreshold) const
{
        _timeJumpThreshold = timeJumpThreshold;
} // OalStateable::timeJumpThreshold()


//
//	---( OalServer )----------------------------------------------------------
//
//	--- constructors ---
//
void OalServer::init() const
{
	_odf = 0;
	_timeCorrelator = 0;
	_attitudeLocator = 0;
	_positionLocator = 0;
	_instrumentalTimer = 0;
} // OalServer::init()


OalServer::OalServer(Xmm::Instrument instrument,
					 Xmm::DataMode dataMode,
					 unsigned exposureIndex,
                                         bool scheduled,
					 Ccd ccd,
					 Node node,
					 bool attitudeFromStr,
					 const string &exposureId,
                                         unsigned timeJumpThreshold)
	: OalStateable(instrument, dataMode, exposureIndex, scheduled,
				   ccd, node, attitudeFromStr,
                                   exposureId, timeJumpThreshold)
{
	init();
} // OalServer::OalServer()


OalServer::OalServer(const OalStateable &oalStateable)
			: OalStateable(oalStateable)
{
	init();
} // OalServer::OalServer()


void OalServer::setState(const string &scienceFileName)
{
	*this<<scienceFileName;
} // OalServer::setState()


void OalServer::setStateFromSetName(const string &scienceSetName)
{
	checkOdf();
	const OdfSet *odfSet = _odf->selectSet(scienceSetName);
	instrument(odfSet->instrument());
	dataMode(odfSet->dataMode());
	scheduled(odfSet->scheduled());
	const Exposure *exp = odf()->exposure(instrument(), dataMode(),
										  odfSet->exposureNr(),
										  Odf::EXP_ODF, true,
										  scheduled());
	exposureIndex(exp->index()+1);
	ccd(odfSet->ccdNr());
	node(Node(odfSet->nodeNr()));
} // OalServer::setStateFromSetName()


void OalServer::checkOdf() const
{
	if (!_odf)
		//	no ODF defined yet - instantiate one first
		newOdf();
} // OalServer::checkOdf()


const Odf *OalServer::newOdf(const string &dirName) const
{
	delete _odf;
	delete _timeCorrelator;	_timeCorrelator = 0;
	delete _attitudeLocator; _attitudeLocator = 0;
	delete _positionLocator; _positionLocator = 0;
	delete _instrumentalTimer; _instrumentalTimer = 0;

	_odf = new Odf(dirName, const_cast<OalServer *>(this));
//
//	The following generates too much diagnostic output for the errstr
//	to handle - disable for the time being
//	if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
//		errstr<<"Successfully openened ODF in directory `"<<dirName<<"': "
//			  <<*_odf
//			  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"Successfully openened ODF in directory `"<<dirName<<"'"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	return _odf;
} // OalServer::newOdf()


const Odf *OalServer::odf() const
{
	checkOdf();

	return _odf;
} // OalServer::odf()


static const char *SAS_OBT_MET_FIT = "SAS_OBT_MET_FIT";
									//
									//	name of environment variable
									//	containing parameters controlling
									//	the OBT-MET fit process:
									//
									//		meth[,deg[,rms[,frac,[igngshov]]
									//
									//	with:
									//		meth: 0/1 (toggle switch to enforce
									//			  recalculation of propagation
									//			  delays)
									//		deg	: maximum degree of
									//			  fit polynomial
									//		rms	: root-mean-square fit
									//			  abandoning citerion
									//		frac: maximum fraction of
									//			  data points to be removed
									//			  in finding good fit
									//		igngshov:
									//			  0/1 toggle switch: if set
									//			  to 1, GS handovers are
									//			  ignored, i.e. the data
									//			  from all GS are taken into
									//			  account; if set to 0 all
									//			  TC after a GS handover
									//			  discarded
									//

static const char *SAS_TIMECORR = "SAS_TIMECORR";
//
//	name of environment variable containing parameter controlling
//	the selection of standard or reconstructed time correlation data
//             TCS     - Only use TCS file for Time Correlation service. If not present in ODF raise an error condition
//             TCS:TCX - Use TCS file if present, if not present use TCX file
//             TCX:TCS - Use TCX file if present, if not present use TCS file
//             TCX     - Only use TCX files. If not present raise an error condition
//      N.B parameter 'meth' in SAS_OBT_MET_FIT is ignored if TCX file is used

static const char *SAS_TIMECORR_TCX = "SAS_TIMECORR_TCX";
//
//	name of environment variable containing parameter controlling
//	the selection of time range from reconstructed time correlation data
//             ALL     - Generate a fit of the time correlation based on the entire file
//             OBS     - Generate a fit of the time correlation based on the data corresponding to the Observation boundaries

const TimeCorrelator *OalServer::newTimeCorrelator(const STime &timeBase) const
{
	delete _timeCorrelator;

	// Default to TCX
	TimeCorrelator::TimeCorrSource    tcorrsrc = TimeCorrelator::TCX_THEN_TCS;
	TimeCorrelator::TimeCorrTimerange tcorrtim = TimeCorrelator::OBS;
	TimeCorrelator::ConstructionMethod	method =	TimeCorrelator::TCS_ONLY;
							//	use TCS data only

	short int	deg	= 4;			//	fourth order polynomial
	double		rms	= 5.e-5;		//	fairly stringent RMS
	double		frac	= .9;			//	90% of points can be removed
	bool		igngshov= true;			//	ignore GS handovers

	char *specPtr = ::getenv(SAS_OBT_MET_FIT);
	if (specPtr) {
		vector<string> v(breakup(specPtr, ","));
		if (v.size() < 1 || v.size() > 5)
			errstr<<"Error in decoding `"<<specPtr<<"' - syntax is: "
					"method,[degree[,rms[,frac[,igngshov]]]"
				  <<Sas::error("FormatError");

		unsigned m = toNumber<unsigned>(v[0]);
		if (m>UNT(TimeCorrelator::TCS_RECALCULATE_DELAYS))
			errstr<<"Construction method must be 0 or 1 - got "<<m
				  <<Sas::error("InvalidTimeCorrelatorConstruction");
		method = TimeCorrelator::ConstructionMethod(m);

		if (v.size() > 1)
			deg = toNumber<short int>(v[1]);
			if (deg<-1)
				//	want piecewise-linear interpolation
				deg = -1;

		if (v.size() > 2) {
			rms = toNumber<double>(v[2]);
			if (rms <= 0.)
				errstr<<"RMS of OBT-MET fit must be >=0 - got: "<<rms
					  <<Sas::error("InvalidRMSvalue");
			if (deg<0 && rms != 0) {
				errstr<<"Piecewise-linear interpolation requested - setting "
						"RMS to 0"
					  <<Sas::message(Msg::Lib, Msg::Verbose);
				rms = 0.;
			}
		}

		if (v.size() > 3)
			frac = toNumber<double>(v[3]);
			if (frac < 0. || frac >= 1.)
				errstr<<"Fraction of removable points in OBT-MET fit must be "
						"in interval [0, 1) - got: "<<frac
					  <<Sas::error("InvalidFracvalue");
		if (v.size() > 4)
			igngshov = toNumber<unsigned>(v[4]) != 0;
	}

	//
	//	check if parameters are to be obtained from env
	//
	const char *tcorrSrcPtr = ::getenv(SAS_TIMECORR);
	if (tcorrSrcPtr) {
		vector<string> sv(::breakup(string(tcorrSrcPtr), ":"));
		if (sv.size() < 1 || sv.size() > 2)
		  errstr << "Invalid value of environment variable `"
			 << SAS_TIMECORR << "': "
			 << tcorrSrcPtr
			 << Sas::error("InvalidValue");

		for (unsigned n=0; n<sv.size(); ++n) {
			sv[n] = ::uppercase(sv[n]);
			if (sv[n] != "TCS" && sv[n] != "TCX")
			  errstr << "Invalid value of environment variable `"
				 << SAS_TIMECORR << "': "
				 << tcorrSrcPtr
				 << Sas::error("InvalidValue");
		}

		if (sv.size() == 2 && sv[0][2] == sv[1][2])
		  errstr << "Invalid value of environment variable `"
			 << SAS_TIMECORR << "': "
			 << tcorrSrcPtr
			 << Sas::error("InvalidValue");

		if (sv.size() == 1)
			tcorrsrc = sv[0][2] == 'S' ? TimeCorrelator::TCS : TimeCorrelator::TCX;
		else
			tcorrsrc = sv[0][2] == 'S' ? TimeCorrelator::TCS_THEN_TCX : TimeCorrelator::TCX_THEN_TCS;
	}

	//
	//	check if parameters are to be obtained from env
	//
	const char *tcorrTimPtr = ::getenv(SAS_TIMECORR_TCX);
	if (tcorrTimPtr) {

       		string value = ::uppercase(string(tcorrTimPtr));
		if (value != "OBS" && value != "ALL")
		  errstr << "Invalid value of environment variable `"
			 << SAS_TIMECORR_TCX << "': "
			 << tcorrTimPtr
			 << Sas::error("InvalidValue");

		tcorrtim = value[0] == 'O' ? TimeCorrelator::OBS : TimeCorrelator::ALL;

	}

	errstr<< "Constructing time correlator object with:\n"
	      << "   Data selection                : "<< tcorrsrc << endl
	      << "   File filter                   : "<< (tcorrtim ? "ALL" : "OBS") << endl
	      << "   construction method           : "<< method << endl
	      << "   maximum polynomial degree     : "<< deg << endl
	      << "   RMS of fit                    : "<< rms << endl
	      << "   fraction of removable points  : "<< frac << endl
	      << "   ignore GS handovers in TC data: "<< (igngshov ? "yes" : "no")
	      << Sas::message(Msg::Lib, Msg::Verbose);

	TimeCorrelator::TimeCorrelatorSpecs tSpecs;
	tSpecs.source = tcorrsrc;
	tSpecs.range = tcorrtim;
	tSpecs.method = method;

	TimeCorrelator::DataInterpolatorSpecs obtMetSpecs = { deg, rms, frac, igngshov };
	TimeCorrelator::DataInterpolatorSpecs ertMetSpecs = { deg, rms, frac, igngshov };

	return _timeCorrelator = new TimeCorrelator(this, tSpecs, obtMetSpecs, ertMetSpecs, timeBase);

} // OalServer::newTimeCorrelator()


const TimeCorrelator *OalServer::timeCorrelator() const
{
	return _timeCorrelator ? _timeCorrelator : newTimeCorrelator();
} // OalServer::timeCorrelator()

//
//	name of environment variable determining the logic with which star tracker
//	attitude data (AHF=Attitude History File, RAF=Raw Attitude File) in an ODF
//	is accessed; possible values (see also doc) are:
//
//		AHF		: use only AHF; error is missing
//		AHF:RAF	: use AHF by d/f; if AHF missing use RAF
//		RAF:AHF	: use RAF by d/f; if RAF is missing use AHF
//		RAF		: use only RAF; error if missing
//
static const char *SAS_ATTITUDE = "SAS_ATTITUDE";

const AttitudeLocator *OalServer::newAttitudeLocator(double maxDelta)
const
{
	delete _attitudeLocator;
	//	use AHF by d/f; if missing try RAF
	AttitudeLocator::AttitudeODFSource attsrc = AttitudeLocator::AHF_THEN_RAF;

	#define INVALID	\
		errstr<<"Invalid value of environment variable `" \
			  <<SAS_ATTITUDE<<"': "<<attSrcPtr<<Sas::error("InvalidValue");

	const char *attSrcPtr = ::getenv(SAS_ATTITUDE);
	if (attSrcPtr) {
		vector<string> sv(::breakup(string(attSrcPtr), ":"));
		if (sv.size() < 1 || sv.size() > 2)
			INVALID

		for (unsigned n=0; n<sv.size(); ++n) {
			sv[n] = ::uppercase(sv[n]);
			if (sv[n] != "RAF" && sv[n] != "AHF")
				INVALID
		}

		if (sv.size() == 2 && sv[0][0] == sv[1][0])
			INVALID;

		if (sv.size() == 1)
			attsrc = sv[0][0] == 'A' ? AttitudeLocator::AHF_ONLY
									 : AttitudeLocator::RAF_ONLY;
		else
			attsrc = sv[0][0] == 'A' ? AttitudeLocator::AHF_THEN_RAF
									 : AttitudeLocator::RAF_THEN_AHF;

	}

	#undef INVALID

	_attitudeLocator = new AttitudeLocator(this, attsrc, maxDelta);

	return _attitudeLocator;
} // OalServer::newAttitudeLocator()

const AttitudeLocator *OalServer::attitudeLocator() const
{
	return _attitudeLocator ? _attitudeLocator : newAttitudeLocator(20.);
} // OalServer::attitudeLocator()


const InstrumentalTimer *OalServer::instrumentalTimer() const
{
	if (!_instrumentalTimer ||
		_instrumentalTimer->instrument() != instrument()) {
		delete _instrumentalTimer;
		_instrumentalTimer = new InstrumentalTimer(
				XmmPayLoad::instrument(instrument()).timeResolution(),
									   instrument(), this);
	}

	return _instrumentalTimer;
} // OalServer::instrumentalTimer()


//
//	name of environment variable determining the type of orbit
//	file that is used; possible value
//	'R'	: reconstructed orbit
//	'P'	:- predicted orbit
//	anything else or undefined: reconstructed or predicted (see explanation
//	of SAS_ORBIT in doc
//
static const char *SAS_ORBIT = "SAS_ORBIT";

const XmmPositionLocator *OalServer::newPositionLocator(double maxDelta) const
{
	delete _positionLocator;

	unsigned method = XmmPositionLocator::FROM_RECONSTRUCTED_ORBIT |
					  XmmPositionLocator::FROM_PREDICTED_ORBIT;

	//
	//	check if orbit file switch given in env
	//
	const char *specPtr = ::getenv(SAS_ORBIT);
	if (specPtr) {
		string s(specPtr);
		if (!s.empty() && toupper(s[0]) == 'R')
			method = XmmPositionLocator::FROM_RECONSTRUCTED_ORBIT;
		if (!s.empty() && toupper(s[0]) == 'P')
			method = XmmPositionLocator::FROM_PREDICTED_ORBIT;
	}

	return _positionLocator = new XmmPositionLocator(this, method, maxDelta);
} // OalServer::newPositionLocator()


const XmmPositionLocator *OalServer::positionLocator() const
{
	return _positionLocator ? _positionLocator : newPositionLocator();
} // OalServer::positionLocator()


const FrameCounterVector *OalServer::newFrameCounterVector(
			const int32 counterVec[], unsigned nCounter) const
{
	FrameCounterVector *v = 0;
	switch (instrument()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: v = new InstrFrameCounterVector<EMOScam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::EPN  : v = new InstrFrameCounterVector<EPNcam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::RGS1 :
		case Xmm::RGS2 : v = new InstrFrameCounterVector<RGScam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::OM   : v = new InstrFrameCounterVector<OMcam>(
									counterVec, nCounter, this);
						 break;
		/* AI SOC-SCR-284
		 * ERM time conversion implementation.
		 */
		case Xmm::ERM   : v = new InstrFrameCounterVector<ERMcam>(
									counterVec, nCounter, this);
						 break;
		default:
						 errstr<<"Cannot allocate a frame counter vector "
								 "for instrument `"
							   <<Xmm::toString(instrument())<<"' - that does "
								 "not make much sense, does it?"
							   <<Sas::error("IllegalInstrument");
	}

	return v;
} // OalServer::newFrameCounterVector()


OalServer::~OalServer()
{
	delete _odf;
	delete _timeCorrelator;
	delete _instrumentalTimer;
} // OalServer::~OalServer()


OalServer &operator<<(OalServer &server, const string &dataSet)
{
	string s(dataSet);
	//	see if the file can be opened 'as is'
	ifstream f(s.c_str());
	if (f)
		f.close();
	else
		//	prepend ODF directory name
		s = oalServer.odf()->directoryName() + '/' + s;

	DataSet *set = dataSetServer->open(s, DataSet::Read, DataSetServer::Low);

	//	check if this is an ODF ICD compliant set name
	server<<*set;

	dataSetServer->close(set);

	return server;
} // operator<<


static ostream &statewarn(const string &listOfAttribs,
						  const string &statevar)
{
	errstr<<"In defining the OAL state from a data set or block "
			"(one or more of) the mandatory attribute(s) "
		  <<listOfAttribs
		  <<" was/were not found - from this/these attribute(s) the actual "
		  <<statevar
		  <<" is derived which is now not possible - will "
			"assume a default value for the "<<statevar<<" - please check "
			"the diagnostic output below (requires verbosity >= Verbose) "
			"for the effective Oal state settings. Please note: "
			"In case of a possibly incorrect "<<statevar<<" all OAL "
			"operations which depend on the correct setting of this/these "
			"variable(s) will either fail to produce correct result or "
			"lead to an anomalous program termination. All OAL operations "
			"which do NOT depend on this/these variable(s) shall work "
			"normally."
		  <<Sas::warning("MandatoryAttributeNotFound");
	return errstr;
} // statewarn()


OalServer &operator<<(OalServer &server, const vector<const Attributable *> av)
{
	string modeStrings[] = { // EMOS
							 "IMAGE.EL", "TIME.EL", "RIMAGE.EL", "CTIME.EL",
							 "AUX.EL", "DIAG.IM", "OFFVAR.EL", "COUNT.EL",
							 "PERHK.EL", "HBR1_NPHK", "HBR2_NPHK",
							 "HBR3_NPHK", "EH_NPHK", "TH_NPHK", "BP_NPH",

							 // EPN
							 "BURST.EL", "OFFSET.IM", "NOISE.IM", "DISCLIN.IM",
							 "MAINHK.EL", "ADDHK.EL",

							 // ERM
							 "ERMCOUNT.EL", "SPECTRA.EL",

							 // RGS
							 "SPECT.EL", "HTR.EL", "FULLHK.EL", "TEMPHK.EL",
							 "DPP_HK.EL",

							 // OM
							 "FST.EL",
							 "ENG1.EL", "ENG2.EL", "ENG3.EL",
							 "ENG4.EL", "ENG5.EL", "ENG6.EL", "ENG7.EL",
							 "FAQ.EL", "WDW.EL", "BFAST.EL",
							 "TRH.EL", "REF.EL", "NONHK.EL",

							 //	SC
							 "ATT", "DUMATT", "PORB", "RECORB", "TIMECORR",
							 "HK_P1", "HK_P2", "HK_P3", "HK_P4", "HK_P5",
							 "HK_P6", "HK_P7", "HK_P8", "HK_P9", "SUM", "" };
	StringIntMap<Xmm::DataMode> dataMode(modeStrings);

	//
	//	make sure we have an ODF
	//
	oalServer.checkOdf();
	const Odf *odf = oalServer.odf();

	//
	//	see if we have an XMM data set
	//
	try {
		string s(DalUtils::attribute(av, "TELESCOP")->asString());
		if (s != "XMM")
			errstr<<"The data set used to define the Oal state does not seem "
					"to be an XMM set - the value of the attribute "
					"`TELESCOP\' is: "<<s
				  <<Sas::error("IllegalAttributeValue");
	} catch(...) {
		errstr<<"No attribute named `TELESCOP\' found"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	//
	//	determine the instrument
	//
	Xmm::Instrument instr = Xmm::SC;
	try {
		instr = Xmm::toInstrument(DalUtils::attribute(av,
													"INSTRUME")->asString());
	} catch(...) {
		errstr<<"The data set used to define the Oal state does not contain "
				"an attribute named `INSTRUME\' - this is needed to determine "
				"the current instrument."
			  <<Sas::error("MandatoryAttributeNotFound");
	}
	oalServer.instrument(instr);

	//
	//	determine exposure number/data mode
	//
	unsigned expNr = 1;
	try {
		string expId = DalUtils::attribute(av, "EXP_ID")->asString();
		if (expId.length() == 13)
			expNr = toNumber<unsigned>(expId.substr(10, 3));
		else
			errstr<<"Exposure identifier `"<<expId<<"' is invalid - "
					"state variable exposureNr will retain its default "
					"value"<<Sas::warning("IllegalExposureNumber");
	} catch(...) {
		statewarn("`EXP_ID\'", "exposure ID");
	}

	//	determine if the exposure was scheduled; this information can
	//	only be inferred from the original file name in attribute
	//	FILENAME of the data set
	bool scheduled = true;
	const Attributable *att = av[0];
	if (!att->hasAttribute("FILENAME")) {
		//	this could be a block - get to the data set
		const Block *b = 0;
		try {
			if ((b = dynamic_cast<const Block *>(att)))
				att = b->parent();
		} catch (...) {}
	}

	if (att && att->hasAttribute("FILENAME")) {
		std::string setName =
			Filename::split(att->attribute("FILENAME")->asString()).second;
		if (OdfSet::isValid(setName)) {
			OdfSet odfSet(setName);
			scheduled = odfSet.scheduled();
		} else
			errstr<<"`"<<setName<<"' is not an ODF ICD compliant ODF set "
					"name - cannot obtain scheduled/unscheduled flag - "
					"will continue assuming exposure was scheduled"
				  <<Sas::warning("NoOdfSetName");
	} else
		errstr<<"Attribute `FILENAME' not found cannot obtain "
				"scheduled/unscheduled flag - will continue assuming "
				"exposure was scheduled"
			  <<Sas::warning("NoOdfSetName");

	oalServer.scheduled(scheduled);
	if (expNr > 0) {
		Xmm::DataMode mode = odf->whatMode(instr, expNr, scheduled);
		if (mode == Xmm::DATA_MODE_NONE) {
			//
			//	try to get data mode from DATATYPE attribute
			//
			string dataType;
			try {
				dataType = DalUtils::attribute(av, "DATATYPE")->asString();
			} catch(...) {
				statewarn("`DATATYPE\'", "data mode");
			}
			Xmm::DataMode mode = dataMode[dataType];
			if (mode == Xmm::DATA_MODE_NONE)
				errstr<<"`"<<dataType<<"' is not a recognized data mode name -"
					  <<" datatype will retain its default setting"
					  <<Sas::warning("UnrecognizedValue");
		}
		oalServer.dataMode(mode);
		const Exposure *exp = oalServer.odf()->findExposure(instr, mode, expNr,
															scheduled);
		assert(exp);
		oalServer.exposureIndex(exp->index()+1);
	} else
		errstr<<"Exposure number must be greater than 0 - state variable "
				"expososurNr will retain its default value"
			  <<Sas::warning("IllegalExposureNumber");

	//
	//	determine CCD/node number
	//
	switch(instr) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			try {
				oalServer.ccd(DalUtils::attribute(av, "CCDID")->asInt());
			} catch(...) {
				statewarn("`CCDID\'", "CCD");
			}
			try {
				oalServer.node(Xmm::toNode(DalUtils::attribute(av,
														"CCDNODE")->asInt()));
			} catch(...) {
				statewarn("`CCDNODE'", "node");
			}
			break;
		case Xmm::EPN:
			try {
				oalServer.ccd(EPNcam::toChipNr(
								DalUtils::attribute(av, "CCDID")->asInt(),
								DalUtils::attribute(av, "QUADRANT")->asInt()));
			} catch(...) {
				statewarn("`CCDID\', `QUADRANT\'", "CCD");
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			try {
				oalServer.ccd(DalUtils::attribute(av, "CCDID")->asInt());
			} catch (...) {
				statewarn("`CCDID\'", "CCD");
			}
			break;
		case Xmm::OM:
			oalServer.ccd(1);	//	lovely, there is only one CCD
			break;
		default:
			errstr<<"Initialization from `"<<Xmm::toString(instr)<<"' not "
					"supported"<<Sas::error("IllegalInstrument");
			break;
	}

	//
	//	all done, output distagnostic messges
	//
	errstr<<"OAL state set to:\n"
		  <<"   instrument             : "<<instr<<endl
		  <<"   data mode              : "<<oalServer.dataMode()<<endl
		  <<"   exposure index         : "<<oalServer.exposureIndex()<<endl
		  <<"   exposure number        : "<<oalServer.exposureNumber()<<endl
		  <<"   scheduled              : "<<(oalServer.scheduled() ?
												"true" : "false")<<endl
		  <<"   ccd number             : "<<oalServer.ccd()<<endl
          <<"   node number            : "<<oalServer.node()<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	return server;
}


OalServer &operator<<(OalServer &server, const Block &block)
{
	errstr<<"Setting OAL state from block `"<<block.qualifiedName()
		  <<"'"<<Sas::message(Msg::Lib, Msg::Verbose);
	vector<const Attributable *> av;
	av.push_back(&block);

	return server<<av;
} // OalServer::operator<<


OalServer &operator<<(OalServer &server, const DataSet &dataSet)
{
	errstr<<"Setting OAL state from data set `"<<dataSet.qualifiedName()
		  <<"'"<<Sas::message(Msg::Lib, Msg::Verbose);
	vector<const Attributable *> av;
	av.push_back(&dataSet);
	av.push_back(dataSet.block(0));

	return server<<av;
} // OalServer::operator<<


DataSet &operator<<(DataSet &set, const OalServer &server)
{
	ExceptionErrorHandler local;
	ErrorHandler *oldHandler = errHandler.handler(&local);

	#define ADD_ATT(code)												\
	try {																\
		code;															\
	} catch (SasException::Exception e) {								\
		errHandler.handler(oldHandler);									\
		errstr<<e.message<<" - no new attributes will be added to "		\
				"data set"<<Sas::warning("NoAttributesAdded");			\
	}

	ADD_ATT(server.attitudeLocator()->addCommonAttributes(&set))
	ADD_ATT(server.positionLocator()->addCommonAttributes(&set));
	ADD_ATT(server.timeCorrelator()->addCommonAttributes(&set));

	errHandler.handler(oldHandler);

	return set;
} // OalServer::operator<<

