//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrelator.cc	--- implementation of TimeCorrelator class
//
//	DESCRIPTION:
//		Implementation part of TimeCorrelator class
//
//	CHANGE HISTORY:
//		UL, 1998-07-16: 1.0.0: - first version
//		UL, 1999-03-04: 1.1.0: - necessary changes related to lowmem Dal
//		UL, 1999-04-28: 1.2.0: - changed fatal -> error
//		UL, 1999-08-20: 1.3.0: - added more diagnostic messages
//		UL, 2000-01-11: 1.4.0: - brought in line with ODF ICD issue 2.4
//								 (FTFINE column of TCS is now unsigned)
//		UL, 2001-03-02: 1.5.0: - implemented recalculation of signal
//								 propagation delays (using column ERTIMRAW
//								 in TCS); option to switch to either scheme
//								 via environment variable SAS_OBT_MET_FIT
//							   - bug fix: TCS file was not closed
//		UL, 2001-03-30: 1.3.0: - added
//									TimeCorrelator::toOnBoardTime()
//							   - give some routines better names
//		UL, 2001-04-04: 1.4.0: - construct MET-OBT fit polynomial
//								 with original OBT-MET poly specs
//		UL, 2001-05-31: 1.4.1: - ignore entries with UTCDAY==0 in TCS
//		UL, 2001-07-24: 1.5.0: - ignore also entries with UTCFLAG!=0 and
//								 all entries after an apparent OBT reset
//								 (see comments below); fixes problem of
//								 SPR 2127
//		UL, 2001-12-10: 1.6.0: - merged two TimeCorrelator construction
//								 methods into one
//		UL, 2002-01-03: 1.7.0: - make robust against non-numeric GSID values
//								 from TCS
//							   - added TimeCorrelator::constructionMethod()
//		UL, 2002-03-25: 1.8.0: - if number of TC data points is now less than
//								 5 times the requested degree of polynomial
//								 revert back to linear fit
//							   - warning if OBT-MET polynomial is evaluated
//								 outsise abscissa data validity range
//		UL, 2003-02-07: 1.9.0: - implemented piecewise-linear interpolation
//								 to TC data (switched own by assigning negative
//								 value to degree-field in SAS_OBT_MET_FIT;
//								 various code changes+reorganizations related
//								 to this
//		UL, 2003-06-27: 1.10.0:- re-calculation of propagation delays:
//								 it was confirmed that OB-delays have to be
//								 SUBTRACTED; corrected that correspondingly
//
//	AUTHOR:
//		UL, Thu Jul 16 10:40:31 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/oal/src/TimeCorrelator.cc,v 1.46 2014/04/15 15:54:31 aibarra Exp $
//
#include <iomanip>
#include <algorithm>

#include "TimeCorrelator.h"
#include "CalOalUtility.h"
#include "OalServer.h"
#include "OalDalUtils.h"
#include "PhysicalConstants.h"
#include "Numeric.h"
#include "LinearInterpolant.h"
#include "Dal.h"
#include "Tracer.h"
#include "slatec.h"
#include "errstr.h"

using namespace CalOalUtility;
using namespace std;

//
//	---------------------( FitPolynomial )-----------------------------------
//
class FitPolynomial : public TimeCorrelationDataInterpolator {
public:
	FitPolynomial(const vector<double> &x,
				  const vector<double> &y,
				  const short int maxDegree, double rms)
		: TimeCorrelationDataInterpolator(x, y, maxDegree, rms)
	{
		if (rms <= 0. || maxDegree == 0)
			errstr<<"RMS must be >= 0 and maximum allowed degree of fit "
					"polynomial must be >0"<<Sas::error("IllegalValues");

		int nPoints = x.size();
		if (nPoints > maxDegree) {
			double	*weights = new double [nPoints],
					*vals = new double [nPoints],
					*xp = new double [nPoints],
					*yp = new double [nPoints];
			for (int n=0; n<nPoints; ++n) {
				xp[n] = x[n]; yp[n] = y[n];
			}
			_fitPoly	= new double [3*nPoints+3*maxDegree+3];
			weights[0]	= -1.;		// do an unweighted least-square-fit
			_rms		= 1.e-20;	// polynomal fit will not achieve this RMS,
									// thus, we will always use the maximum
									// possible degree
			int ierr, degree;		//	error code from SLATEC

		    SLATEC_DPOLFT(nPoints, xp, yp, weights, maxDegree, degree, _rms,
						  vals, ierr, _fitPoly);
			_degree = degree;

			if (ierr != 1 && ierr != 3)
				//
				//	fit failed
				//
				errstr<<"Routine `DPOLFIT' failed to find a polynomal fit of "
						"maximum degree "<<maxDegree
					  <<" - return code is "<<ierr
					  <<Sas::error("UnsuccessfulFit");

			//
			//	fit is ok
		//
			errstr<<"Fit polynomial computed - optimized degree: "<<_degree
				  <<" - RMS error: "<<_rms
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			delete [] weights;
			delete [] vals;
			delete [] xp;
			delete [] yp;
		} else
			errstr<<"Only "<<nPoints<<" data points are available - cannot do "
					"a fit to a polynomial of degree "<<maxDegree
				  <<Sas::error("TooFewDataPoints");
	} // FitPolynomial()


	double evaluate(double x, bool warningOnExtrapolation = true) const
	{
		double retVal;

		if (warningOnExtrapolation && (x<_x.front()-100. || x>_x.back()+100.))
			//	issue extrapolation warning
			errstr<<"Evaluating OBT-MET fit polynomial at x="<<setprecision(10)
				  <<x<<" which is outside the time correlation data abscissa "
					"range ["<<setprecision(10)<<_x.front()<<", "<<_x.back()
				  <<"] - result of extrapolation might be afflicted with "
					"large numerical uncertainty."
				  <<Sas::warning("Extrapolation");

		SLATEC_DP1VLU(_degree, 0, x, retVal, static_cast<double *>(0),
					  _fitPoly);

		return retVal;
	} // evaluate()


	double derivative(double x, unsigned short n,
					  bool warningOnExtrapolation) const
	{
		if (n==0)
			return evaluate(x, warningOnExtrapolation);

		double retVal;
		double deriv[n];
		int di = INT(n);

		SLATEC_DP1VLU(_degree, di, x, retVal, deriv, _fitPoly);

		return deriv[n-1];
	} // derivative()


	~FitPolynomial()
	{
		delete [] _fitPoly;
	} // ~FitPolynomial()

private:
	double		*_fitPoly;	// the result of the fit is a populated double
							// vector defining the polynomal - will be used
							// by evaluate() to compute polynomial at
							// a certain point
};


//
//	---------------------( PiecewiseLinearInterpolator )-----------------------
//
class PiecewiseLinearInterpolator : public TimeCorrelationDataInterpolator,
									public Slatec::LinearInterpolant<double> {
public:
	PiecewiseLinearInterpolator(const vector<double> &x,
								const vector<double> &y)
		: Slatec::Interpolant<double>(x),
		  TimeCorrelationDataInterpolator(x, y, -1, 0.),
		  Slatec::LinearInterpolant<double>(x, y) {}

	double evaluate(double x, bool = true) const
	{
		return operator()(x);
	} // evaluate()


	double derivative(double x, unsigned short n,
					  bool warningOnExtrapolation) const
	{
		if (n==0)
			return evaluate(x, warningOnExtrapolation);

		return 0.; // not implemented
	} // derivative()

	~PiecewiseLinearInterpolator() {}
};


//
// -----------------------------( LinearObtMet )-----------------------------
//

//
//	info class which encapsulates access to environment variable
//	SAS_OBT_MET_LINEAR; if defined it enforces a linear fit the to
//	time correlation (TC) data using a given slope (first field) and
//	a single data point (nrows*fraction (second field) from the
//	TCS
//
class LinearObtMet {
public:
	LinearObtMet() : _enabled(false), _slope(1.), _fraction(0.) {
		char *specPtr;
		if ((specPtr = getenv("SAS_OBT_MET_LINEAR"))) {
			_enabled = true;
			vector<string> v(breakup(specPtr, ","));
			if (v.size() < 1 || v.size() > 2)
				errstr<<"Error in decoding `"<<specPtr<<"' - syntax is: "
						"slope,[fraction]"
					  <<Sas::error("FormatError");

			_slope = toNumber<double>(v[0]);
			_fraction = toNumber<double>(v[1]);
			if (_fraction<0. || _fraction>1.)
				errstr<<"Fraction fields in `"<<specPtr<<"' must lie in "
						"interval [0, 1]"
					  <<Sas::error("ValueOutOfValidBounds");
		}
	}

	bool enabled() const { return _enabled; }
	double slope() const { return _slope; }
	double fraction() const { return _fraction; }

	~LinearObtMet() {}

private:
	bool	_enabled;
	double	_slope, _fraction;
};


//
// ---------------------------( TimeCorrelator )-----------------------------
//
const TimeCorrelationDataInterpolator *
			TimeCorrelator::constructInterpolator(
								DataInterpolatorSpecs &dataInterpolatorSpecs,
								vector<double> &x,
								vector<double> &y) const
{
	if (dataInterpolatorSpecs.degree < 0) {
		//
		//	instantiate piecewise linear interpolator
		//
		errstr<<"Instantiating piecewise-linear interpolator for "
				"time correlation data"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		return new PiecewiseLinearInterpolator(x, y);
	}

	//
	//	otherwise we want a polynomial fit
	//
	int nPoints = x.size(), pointsRemoved = 0;

	//	require that number of abscissa points is at least 5 times the
	//	requested max. degree of fit polynomial - if not revert to
	//	linear fit and issue warning
	if (dataInterpolatorSpecs.degree > 1 &&
		nPoints < 5*dataInterpolatorSpecs.degree) {
		errstr<<"Number of time correlation data points ("<<nPoints
			  <<") is less than 5 times the requested maximum degree of the "
				"OBT-MET fit polynomial ("<<dataInterpolatorSpecs.degree
			  <<") - reverting to linear fit with RMS=1.e-2"
			  <<Sas::warning("TooFewTimeCorrelationDataPoints");
		dataInterpolatorSpecs.degree = 1;
		dataInterpolatorSpecs.rms = 1.e-2;
	}

	FitPolynomial *fitPoly = new FitPolynomial(x, y,
											   dataInterpolatorSpecs.degree,
											   dataInterpolatorSpecs.rms);
	double removedFrac = 0.;
	while (x.size() > UNT(dataInterpolatorSpecs.degree) &&
		   removedFrac <= dataInterpolatorSpecs.discardPoints &&
		   dataInterpolatorSpecs.rms < fitPoly->rms()) {
		//
		//	no satisfactory fit found yet - delete fit polynomial,
		//	remove "worst" data point and re-calculate
		//
		vector<double> diff(x.size());
		for (unsigned n=0; n<x.size(); ++n)
			diff[n] = fabs(y[n] - fitPoly->evaluate(x[n]));
		vector<double>::iterator pos = max_element(diff.begin(), diff.end());
		x.erase(pos-diff.begin()+x.begin());
		y.erase(pos-diff.begin()+y.begin());
		removedFrac = DBL(++pointsRemoved)/DBL(nPoints);

		delete fitPoly;
		fitPoly = new FitPolynomial(x, y,
									dataInterpolatorSpecs.degree,
									dataInterpolatorSpecs.rms);
	}

	if (removedFrac > dataInterpolatorSpecs.discardPoints)
		errstr<<"No satisfactory fit found after removing "
			  <<setprecision(2)<<100.*removedFrac<<"% of the data points"
			  <<Sas::error("UnsuccessfulFit");

	if (x.size() <= UNT(dataInterpolatorSpecs.degree) ||
		dataInterpolatorSpecs.rms < fitPoly->rms())
		errstr<<"Could not find a satisfactory fit"
			  <<Sas::error("UnsuccessfulFit");

	//
	//	be informative
	//
	vector<double> diff(x.size());
	for (unsigned n=0; n<x.size(); ++n)
		diff[n] = y[n] - fitPoly->evaluate(x[n]);

	errstr<<"Fit polynomial found - degree: "<<fitPoly->degree()
		  <<" RMS: "<<setprecision(5)<<fitPoly->rms()
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"Differences between data points and fit polynomial:";
	for (unsigned n=0; n<diff.size(); ++n)
		errstr<<"   "<<setprecision(16)<<x[n]<<" "<<setprecision(16)<<diff[n]
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	dataInterpolatorSpecs.degree = fitPoly->degree();
	dataInterpolatorSpecs.rms = fitPoly->rms();
	dataInterpolatorSpecs.discardPoints = removedFrac;

	return fitPoly;
} // TimeCorrelator::constructInterpolator()

//JSH
void TimeCorrelator::construct(const string &timeCorrelationFileName)
{

	errstr << "Opening Time Correlation file `"
	       << timeCorrelationFileName
	       << Sas::message(Msg::Lib, Msg::Verbose);

	DataSet *set = dataSetServer->open(timeCorrelationFileName, DataSet::Read);
	Table *table = set->table(0);
	unsigned nRows = table->rows();

	//
	//	check if we have column ERTIMRAW - if not this is an old-style
	//	TCS produced before Dec 2000
	//
	if ((timeCorrelatorSpecs().method == TCS_RECALCULATE_DELAYS)  && !table->hasColumn("ERTIMRAW")) {
		errstr<<"Table `"<<table->qualifiedName()<<"' has no column ERTIMRAW "
				"which is needed for the recalculation of time correlation "
				"data - will continue with XMCS generated values"
			  <<Sas::warning("NoSuchColumn");
		//timeCorrelatorSpecs().method = TCS_ONLY;
		//gcc4.8.2 compatibility. New method "setMethos" created.
		timeCorrelatorSpecs().setMethod(TCS_ONLY);
	}

	ColumnData *erttimecol = table->column("ERTIME")->data(),
			   *cdobtcoarse	= table->column("OBTCOARS")->data(),
			   *cdobtfine = table->column("OBTFINE")->data(),
			   *cdutcday = table->column("UTCDAY")->data(),
			   *cdutcmilli = table->column("UTCMILLI")->data(),
			   *cdutcmicro = table->column("UTCMICRO")->data(),
			   *cdutcflag = table->column("UTCFLAG")->data(),
			   *cdgsid = table->column("GSID")->data(),
			   *cdertraw = (timeCorrelatorSpecs().method == TCS_RECALCULATE_DELAYS)  ? table->column("ERTIMRAW")->data(): 0;

	int32	*obtcoarse	= cdobtcoarse->int32Data();
	uint16	*obtfine	= cdobtfine->uint16Data();
	int16	*utcday		= cdutcday->int16Data();
	int32	*utcmilli	= cdutcmilli->int32Data();
	int16	*utcmicro	= cdutcmicro->int16Data(),
		*utcflag	= cdutcflag->int16Data();

	vector<double>	obt, met, ert;

	bool firstRecord = false;

	//
	//	check that we have at least two rows and get inital GS ID
	//
	if (!nRows)
		errstr<<"The Time Correlation data set `"<<set->name()
			  <<"' in the current ODF is empty"
			  <<Sas::fatal("NoTimeCorrelationData");

	if (timeCorrelatorSpecs().range == OBS) {
	  errstr << "Using time range: "
		 << timeCorrelatorSpecs().start.toString()
		 << " to "
		 << timeCorrelatorSpecs().end.toString()
		 << Sas::message(Msg::Lib, Msg::Verbose);
	}


	string gsid0 = stringColumnRow(cdgsid, 0);

	unsigned lastObtCoarse = 0, nWraps = 0, nskips = 0;
	unsigned dataRows = 0;

	for (unsigned n=0; n<nRows; ++n) {

	  STime thisRec = STime(stringColumnRow(erttimecol, n));

	  if ((timeCorrelatorSpecs().source == TCX) && (timeCorrelatorSpecs().range == OBS) && (thisRec < timeCorrelatorSpecs().start)) {
	    errstr << "Skipping record at "
		   << thisRec.toString()
		   << Sas::message(Msg::Lib, Msg::Verbose);
	    continue;
	  }

	  if ((timeCorrelatorSpecs().source == TCX) && (timeCorrelatorSpecs().range == OBS) && (thisRec > timeCorrelatorSpecs().end)) {
	    errstr << "Reached the end of Observation @ "
		   << thisRec.toString()
		   << ", nothing more to do."
		   << Sas::message(Msg::Lib, Msg::Verbose);
	    break;
	  }

	  if (!firstRecord) {

	    if ((timeCorrelatorSpecs().source == TCS) ||
		((timeCorrelatorSpecs().source == TCX) && (timeCorrelatorSpecs().range == ALL)) ||
		((timeCorrelatorSpecs().source == TCX) && (timeCorrelatorSpecs().range == OBS) && (thisRec >= timeCorrelatorSpecs().start))) {

	      firstRecord = true;

	      errstr << "Processing first record at "
		     << thisRec.toString()
		     << Sas::message(Msg::Lib, Msg::Verbose);

	      //
	      //	get very first UTC field and establish _utcBaseTime
	      //
	      _utcBaseTime = STime(1958, 1, 1 + utcday[n]);
	      _utcOffset = _utcBaseTime - _baseTime;

	      errstr<< "TimeCorrelator::utcBaseTime = "
		    << _utcBaseTime
		    << Sas::message(Msg::Lib, Msg::Verbose);
	      errstr<< "TimeCorrelator::baseTime    = "
		    << _baseTime
		    << Sas::message(Msg::Lib, Msg::Verbose);
	      errstr<< "TimeCorrelator::utcOffset   = " << setprecision(12)
		    << _utcOffset
		    << Sas::message(Msg::Lib, Msg::Verbose);

	      lastObtCoarse = OalDalUtils::unsigned32(obtcoarse[n]);

	    }

	  }


	  if (utcday[n] <= 0 || utcflag[n]) {
	    nskips++;
	    continue;	//	ignore entries with
	                //		* UTCDAY==0
                        //		* UTCFLAG!=0
  	                //	note: some early ODFs e.g. 0078/0124100101 have
	                //	UTCDAY==0 but UTCFLAG==0 so, UTCFLAG is not
	                //	a reliable quality indicator
	  }

	  string gsid = stringColumnRow(cdgsid, n);
	  if (gsid != gsid0)
	    gsid0 = gsid;

	  //
	  //	change in GS - see if we want to ignore the remaining data
	  //
	  if (!_obtMetDataInterpolatorSpecs.ignoreGSHandovers)
	    break;

	  //
	  //	get OBT value
	  //
	  unsigned uobtcoarse = OalDalUtils::unsigned32(obtcoarse[n]);
	  double diff = INT(uobtcoarse) - INT(lastObtCoarse);
	  if (diff < 0.) // AI gcc4 porting: suggest explicit braces to avoid ambiguous 'else'
	  {
	    if (-diff > .9*65536.*65536.)
	      //
	      //	coarse time must have had a wrap around
	      //
	      ++nWraps;
	    else {
	      //
	      //	this is an OBT reset (e.g. ODF 102_0099020301 jump
	      //	from 1114402 to 33 at TCS record 2827) - cannot use
	      //	subsequent time	correlation entries; ignore the rest
	      //
	      errstr <<"Detected apparent reset of onboard clock in time correlation data table `"
		     << table->qualifiedName()
		     << " between records #"
		     << n << '/' << n+1
		     << " where OBTCOARSE jumps from "
		     << lastObtCoarse << " to " << uobtcoarse << " - ignoring all "
		"subsequent records. The impact of this on "
		"the accuracy of event arrival times is very likely "
		"to be negligible. If you intend to perform "
		"high-accuracy timing analysis on this ODF data set, "
		"please consult the XMM-Newton helpdesk and ask for "
		"confirmation of this assessment. In doing so please "
		"detail the complete ODF designation and the version "
		"of the SAS you are using at this moment."
		     <<Sas::warning("OBTreset");
	      break;
	    }
	  }

	  lastObtCoarse = uobtcoarse;

	  obt.push_back(DBL(uobtcoarse) +
			DBL(OalDalUtils::unsigned32(obtfine[n]))/65536. +
			DBL(nWraps) * 65536. * 65536.);

	  //
	  //	get UTC value
	  //
	  if (timeCorrelatorSpecs().source == TCX)
	    met.push_back(stringColumnRow(erttimecol, n) - _utcBaseTime );
	  else {
	    if (timeCorrelatorSpecs().method == TCS_ONLY)
	      met.push_back(STime(1958, 1, 1+utcday[n], 0, 0, 0,
				  1.e-3 * DBL(utcmilli[n]) + 1.e-6 * DBL(utcmicro[n])) - _utcBaseTime);
	    else
	      met.push_back(metWithPropDelays(stringColumnRow(cdertraw, n), gsid));
	  }


	  //
	  //	get ERT value
	  //
	  ert.push_back(STime(stringColumnRow(erttimecol, n)) - _utcBaseTime);

	  errstr << "TimeCorrDataPoint: "
		 << setprecision(12)
		 << obt[dataRows]
		 << " "
		 << met[dataRows]
		 << " "
		 << ert[dataRows]
		 << " "
		 << obt[dataRows]-obt[0]
		 << " "
		 << met[dataRows]-met[0]
		 << Sas::message(Msg::Lib, Msg::Verbose);

	  dataRows++;

	}

	//	tidy up and close TCS
	delete erttimecol;
	delete cdertraw;
	delete cdobtcoarse;
	delete cdobtfine;
	delete cdutcday;
	delete cdutcmilli;
	delete cdutcmicro;
	delete cdutcflag;
	delete cdgsid;

	dataSetServer->close(set);

	//
	//	Do we want to enforce a linear OBT-MET relation?
	//
	LinearObtMet lininfo;
	if (lininfo.enabled() || obt.size() == 1) {
		double slope = 1.;
		if (lininfo.enabled()) {
			slope = lininfo.slope();
			unsigned nPt = UNT(obt.size()*lininfo.fraction());
			double obt0 = obt[nPt], met0 = met[nPt], ert0 = met[nPt];
			obt.clear(); obt.push_back(obt0);
			met.clear(); met.push_back(met0);
			ert.clear(); ert.push_back(ert0);
		} else
			errstr<<"Only one usable time correlation data point in TCS "
					"of current ODF - will assume perfect linear relationship "
					"between OBT and MET"
				  <<Sas::warning("AssumedOBT_METrelation");

		obt.push_back(obt.back()+1.);
		ert.push_back(ert.back()+1.);
		met.push_back(met.back()+slope);
	}

	//
	//	construct data interpolator
	//
	vector<double> met2(met);
	DataInterpolatorSpecs obtMetDataInterpolatorSpecs = _obtMetDataInterpolatorSpecs;
	_obtMetInterpolator =
			constructInterpolator(_obtMetDataInterpolatorSpecs, obt, met);
	_metObtInterpolator =
			constructInterpolator(obtMetDataInterpolatorSpecs, met, obt);
	_ertMetInterpolator =
			constructInterpolator(_ertMetDataInterpolatorSpecs, ert, met2);

	errstr<<"OBT-MET data interpolator successfully calculated - signature:\n"
		  <<*_obtMetInterpolator
		  <<Sas::message(Msg::Lib, Msg::Sparse);
} // TimeCorrelator::construct()


double TimeCorrelator::metWithPropDelays(const string &ertraw,
										 const string &gsid)
{
	static const PositionLocator *posloc = 0;

	if (!posloc)
		posloc = oalServer.positionLocator();

	unsigned short gsi = 0;
	try {
		gsi = toNumber<unsigned short>(gsid);
	} catch(...) {
		errstr<<"Cannot convert GSID `"<<gsid
			  <<"' to valid numeric ground station identifier"
			  <<Sas::error("InvalidGSID");
	}

	const XmmGroundStation &gs = XmmMission::groundStation(gsi);
	STime t(ertraw);
	CartesianVector xmmPos = posloc->stateVector(t).position,
					gsPos = gs.stateVector(t).position;
	double met,
		   xmmGSdist = (gsPos - xmmPos).norm() * 1000.,	//	km -> m
		   propDelay =
				//	signal travel time - NOTE: atmospheric propagation
				//	delays are of the order of 0.1 us (see email
				//	F. Dreger dated 2001-02-28), so, negligible
			    xmmGSdist/PhysicalConstants::c +

				// Ground station delays
				gs.timingDelay() -

				//	OBDH delay - yes, it has to be SUBTRACTED
				XmmMission::dataHandler().timingDelay();

	errstr<<"Total propagation delay at T="<<t.toString(6)
		  <<" (GS="<<gs.earthObject().name()<<") : "<<propDelay<<" s"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	met = t - _utcBaseTime;
	met -= propDelay;

	return met;
} // TimeCorrelator::metWithPropDelays()


TimeCorrelator::TimeCorrelator(const STime &timeBase)
	: OalChild<OalServer>(0), _baseTime(timeBase), _lastObtTag(0.),
							  _obtWraps(0), _obtMetInterpolator(0),
							  _ertMetInterpolator(0),
							  _metObtInterpolator(0)
{
	_utcBaseTime = STime::earliest;

	_obtMetDataInterpolatorSpecs.degree = 0;
	_obtMetDataInterpolatorSpecs.rms = 0.;
	_obtMetDataInterpolatorSpecs.discardPoints = 0;

	_ertMetDataInterpolatorSpecs = _obtMetDataInterpolatorSpecs;
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(	const OdfSet *timeCorrelationSet,
				      TimeCorrelatorSpecs &specs,
				const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
				const STime &timeBase)
	: OalChild<OalServer>(0),
	  _timeCorrelatorSpecs(specs),
	  _obtMetDataInterpolatorSpecs(obtMetDataInterpolatorSpecs),
	  _ertMetDataInterpolatorSpecs(ertMetDataInterpolatorSpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{

  // No OalEngine, so can't get the time range of the observation, force, full rev
  _timeCorrelatorSpecs.range = ALL;
  construct(timeCorrelationSet->name());

} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(	const string &timeCorrelationSetName,
	                              TimeCorrelatorSpecs &specs,
				const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
	                        const STime &timeBase)
	: OalChild<OalServer>(0),
	  _timeCorrelatorSpecs(specs),
	  _obtMetDataInterpolatorSpecs(obtMetDataInterpolatorSpecs),
	  _ertMetDataInterpolatorSpecs(ertMetDataInterpolatorSpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{

  // No OalEngine, so can't get the time range of the observation, force, full rev
  _timeCorrelatorSpecs.range = ALL;
  construct(timeCorrelationSetName);

} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(	const OalServer *oalEngine,
				TimeCorrelatorSpecs &specs,
				const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
				const STime &timeBase)
	: OalChild<OalServer>(oalEngine),
	  _timeCorrelatorSpecs(specs),
	  _obtMetDataInterpolatorSpecs(obtMetDataInterpolatorSpecs),
	  _ertMetDataInterpolatorSpecs(ertMetDataInterpolatorSpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{

	assert(oalEngine);

	const OdfSet *tcsSet = oalEngine->odf()->selectSet(Xmm::SC,
							   Xmm::TIME_CORRELATION, 1,
							   0, Odf::EXP_OAL,
							   false, true, false);
	bool hasTCS = tcsSet;

	const OdfSet *tcxSet = oalEngine->odf()->selectSet(Xmm::SC,
							   Xmm::RECONSTRUCTED_TIME_CORRELATION, 1,
							   0, Odf::EXP_OAL,
							   false, true, false);
	bool hasTCX = tcxSet;

	if (hasTCS && ( _timeCorrelatorSpecs.source == TCS ||
			_timeCorrelatorSpecs.source == TCS_THEN_TCX ||
			(_timeCorrelatorSpecs.source == TCX_THEN_TCS && !hasTCX))) {

	  _timeCorrelatorSpecs.source = TCS;
	  _timeCorrelatorSpecs.range = ALL;
	  construct(tcsSet->name());
	}

	else if (hasTCX && ( _timeCorrelatorSpecs.source == TCX ||
			     _timeCorrelatorSpecs.source == TCX_THEN_TCS ||
			     (_timeCorrelatorSpecs.source == TCS_THEN_TCX && !hasTCS))) {

	  _timeCorrelatorSpecs.source = TCX;

	  if (_timeCorrelatorSpecs.range == OBS) {
	    // Time Correlation chopped on Observation boundaries
	    _timeCorrelatorSpecs.start  = oalEngine->odf()->getDateObsTime(Xmm::SC, Xmm::TIME_CORRELATION);
	    _timeCorrelatorSpecs.end    = oalEngine->odf()->getDateEndTime(Xmm::SC, Xmm::TIME_CORRELATION);
	  }

	  construct(tcxSet->name());
	}
	else
	  errstr << "The ODF in directory `"
		 << oalEngine->odf()->directoryName()
		 << "' does not contain the requested type of time correlation data - TCX and/or TCS missing"
		 << Sas::error("MissingTimeCorrelationData");

} // TimeCorrelator::TimeCorrelator()


void TimeCorrelator::setBaseTime(const STime &) const
{
	errstr<<"TimeCorrelator::setBaseTime is not yet implemented"
		  <<Sas::error("NotYetImplemented");
} // TimeCorrelator::setBaseTime()


double TimeCorrelator::toMissionEllapsedTime(double obtTag) const
{
	if (obtTag < _lastObtTag &&			//	a time decrease
		_lastObtTag - obtTag > 320000.)	//	and difference is large enough
										//	to believe a wrap-around has
										//	occured
		++_obtWraps;

	_lastObtTag = obtTag;
	obtTag += DBL(_obtWraps) * 65536. * 65536.;

	return _obtMetInterpolator->evaluate(obtTag) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(unsigned obtCoarse,
											 unsigned short obtFine) const
{
	return toMissionEllapsedTime(DBL(obtCoarse) + DBL(obtFine)/65536.);
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toOnBoardTime(double missionEllapsedTime) const
{
	return _metObtInterpolator->evaluate(missionEllapsedTime - _utcOffset);
} // TimeCorrelator::toOnBoardTime()


void TimeCorrelator::resetObtMetRelation() const
{
	_lastObtTag = 0.;
	_obtWraps = 0;
} // TimeCorrelator::resetObtMetRelation()


void TimeCorrelator::plotObtMetRelation() const
{
	errstr<<"TimeCorrelator::plotObtMetRelation is not yet implemented"
		  <<Sas::error("NotYetImplemented");
} // TimeCorrelator::plotObtMetRelation()


double TimeCorrelator::toMissionEllapsedTime(const string &ertString) const
{
	return toMissionEllapsedTime(STime(ertString));
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(const STime &ertTag) const
{
	double arg = ertTag - _utcBaseTime;

	return _ertMetInterpolator->evaluate(arg) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


void TimeCorrelator::plotErtMetRelation() const
{
	errstr<<"TimeCorrelator::plotErtMetRelation is not yet implemented"
		  <<Sas::error("Not yet implemented");
} // TimeCorrelator::plotErtMetRelation()

namespace {
	template<class T>
	void addAttrib(Attributable *attrib,
				   const string &name, const T &val, const string &comment) {
		Attribute *att = attrib->hasAttribute(name)
							? attrib->attribute(name)
							: attrib->addAttribute(name, comment);
		*att = val;
	}
}

const TimeCorrelator &TimeCorrelator::addCommonAttributes(Attributable *attrib) const
{
	addAttrib(attrib, "TFIT_RPD",
			  constructionMethod()==TCS_RECALCULATE_DELAYS,
			  "Recalculated signal propagation delays?");
	const DataInterpolatorSpecs &specs = obtMetDataInterpolatorSpecs();
	addAttrib(attrib, "TFIT_DEG", int(specs.degree),
			  "Degree of OBT-MET fit polynomial");
	addAttrib(attrib, "TFIT_RMS", specs.rms,
			  "RMS value of OBT-MET polynomial fit");
	addAttrib(attrib, "TFIT_PFR", specs.discardPoints,
			  "Fraction of disregarded TCS data points");
	addAttrib(attrib, "TFIT_IGH", specs.ignoreGSHandovers,
			  "Ignored GS handover in TC data?");

	return *this;
} // TimeCorrelator::addCommonAttributes()


TimeCorrelator::~TimeCorrelator()
{
	delete _obtMetInterpolator;
	delete _ertMetInterpolator;
	delete _metObtInterpolator;
} // TimeCorrelator::~TimeCorrelator()



ostream &operator<<(ostream &os, const TimeCorrelationDataInterpolator &fp)
{
	short int N=fp.degree();
	os<<"# data points: "<<fp.xPoints().size()<<endl
	  <<"degree       : "<<N<<endl
	  <<"RMS          : "<<setprecision(8)<<fp.rms()<<endl
	  <<"coefficients : "<<endl;
	for (short int n=0; n<=N; ++n)
		os<<"    a"<<n<<" : "
		  <<setprecision(12)<<fp.derivative(0., n, false)/
												Numeric::factorial(n)<<endl;
	return os;
} // operator<<

