//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeTags.cc	--- implementation of TimeTag classes
//
//	DESCRIPTION:
//		Implementation part of classes in Tag.h
//
//	CHANGE HISTORY:
//		UL, 1998-08-03: 1.0.0: - first version
//		UL, 1998-10-14: 1.1.0: - bug fix in toObt() methods
//		UL, 1999-03-04: 1.2.0: - changes to make it work with low memort Dal
//		UL, 1999-03-16: 1.3.0: - changed column name: "EOSCOARSE" -> "EOSCOARS"
//		UL, 1999-04-28: 1.4.0: - fatal -> error
//		UL, 1999-05-25: 1.4.1: - bug fix: was assuming wrong range of
//								 coarse times for all instruments
//							   - frameCounterToObt: do not bail out on
//								 missing frame numbers; just issue
//								 warning message and set OBT to -1.
//							   - modification in frameCounterToObt for EMOS
//		UL, 1999-11-02: 1.4.2: - more informative error message in
//								 frameCounter->obt conversion for RGS
//		UL, 1999-12-17: 1.5.0: - bug fix in
//									InstrFrameCounterVector<EPNcam>::toObt()
//								 was accessing wrong AUX file
//		UL, 1999-12-20: 1.6.0: - work-around for SciSim bug: FTCOARSE for
//								 MOS/PN AUX files in SciSim generated
//								 AUX files is not in range 0..32767 but
//								 0..32767,-32768,-32767...
//		UL, 2000-01-05: 1.7.0: - bug fix in resetTimeEpic(),
//								 fixes:
//								 SPR 957 : Problem with OAL and diagnostic
//										   files on SVT3
//								 SPR 1063 : Behaviour when exposure is not 1.
//		UL, 2000-01-11: 1.8.0: - brought in line with ODF ICD issue 2.4
//								 (FTFINE columns in PN/RGS AUX files are
//								  now unsigned)
//		UL, 2000-02-03: 1.9.0: - fixed SPR  1079 : Unwarranted IllegalObtValue
//								 warning
//								 (emframes subtracts from the OBT value of
//								 the first frame the frame integration time
//								 to estimate the start of the frame. If
//								 the EPIC clock reset time in the PEH
//								 (FTCOARSE/FTFINE) is close to zero this can
//								 then lead to negative OBT times which was
//								 checked against in ObtVector::toUtc(); in
//								 this case the method used to set the time
//								 tag to -1 and issue a warning message; now
//								 it does merely do the latter but computes the
//								 time tag ok).
//		UL, 2000-04-03: 1.10.0:- accessing coarse/fine time components of
//								 EPIC clock reset times in periodic HK
//								 files with right names:
//
//										coarse components	fine component
//											 name/type			 name/type
//								 EMOS1		E1573/J(S)			EU573/I(U)
//								 EMOS2		K1573/J(S)			KU573/I(U)
//								 EPN		F1030/J(S)			FU036/I(U)
//							   - bug fix in frame counter to tag conversions
//								 which falsely caused a wrong-wrap around
//								 detected to be detected
//								 (this fixes SPR 1239 : Wrong wrap around at
//								  first frame)
//		UL, 2000-08-29: 1.11.0:- various changes to read coarse time wrap
//								 around value for EPN clock from periodic
//								 HK parameter F1052 (was assumed to be
//								 constant=32768)
//		UL, 2001-03-27: 1.11.1:- open all HK/AUX data files in lowmem mode
//							   - check the validity of the EPIC clock reset
//								 time/OBT values converted from RGS/OM time
//								 tags by comparison against exposure end
//								 time; if the former lies after the latter
//								 there was a CDMU crash - take the exposure
//								 start time as effective reset time in this
//								 case
//		UL, 2001-07-24: 1.11.2:- fixed minor coding bug
//							   - more tolerant w.r.t missing reset times in HK
//		UL, 2002-03-15: 1.12.0:- ObtVector::toMET(): bug fix: negative OBTs
//								 were correctly converted to NaN but then
//								 overwritten by a meaningless huge positive
//								 number
//		UL, 2002-04-26: 1.13.0:- make use of new HKParamFinder::seek() I/F
//								 to avoid seeking beyond end of exposure
//
//	AUTHOR:
//		U. Lammers, Mon Aug  3 16:10:16 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/oal/src/TimeTags.cc,v 1.24 2013/10/23 17:03:18 aibarra Exp $
//

#include <cassert>
#include <algorithm>

#include "Dal.h"
#include "NAN.h"
#include "Tracer.h"
#include "TimeTags.h"
#include "Xmm.h"
#include "Odf.h"
#include "OalServer.h"
#include "CalOalUtility.h"
#include "OalDalUtils.h"
#include "AuxHKaccess.h"

using namespace CalOalUtility;

//
//	---( EMOS )-------------------------------------------------------------
//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 */
template<> unsigned InstrFrameCounterVector<EMOScam>::_nWraps = 0;
template<> unsigned InstrFrameCounterVector<EMOScam>::_lastWrapAt = 0;
/* AI gcc4 porting
 * undefined reference to `InstrFrameCounterVector<EMOScam>::_resetTimes'
 * => = 0 added
 */
std::vector<double> tmp(0);
template<> std::vector<double> InstrFrameCounterVector<EMOScam>::_resetTimes = tmp;
/* AI gcc4 porting
 * undefined reference to `InstrFrameCounterVector<EMOScam>::_wrapValue'
 * => = 0 added
 */
template<> unsigned InstrFrameCounterVector<EMOScam>::_wrapValue = 0;
template<> unsigned short InstrFrameCounterVector<EMOScam>::_chopperValue = 0;

/* AI 2009-07-08
 * ERM only. Not used for any other instrument
 */
template<> std::string InstrFrameCounterVector<EMOScam>::_auxName="";
template<>
void InstrFrameCounterVector<EMOScam>::setAux(std::string name) const
{
	_auxName = name;
}

template<>
void InstrFrameCounterVector<EMOScam>::setStatics() const
{

	// Check if this is a mosaic poiting observation. If this is the case, we have to set the deltaT value to search
	// for the MOS reset clock values to 1 sec. Otherwise, we can have problems set the time tags correctly for the
	// stable pointing observations
	string prop = parent()->odf()->proposalId();
	int propi = atoi(prop.substr(0,1).c_str());
	string obsid = parent()->odf()->observationId() ;

	int obsi;
	obsi  = atoi(obsid.substr(2).c_str());
	double deltaT = 90.;

	if ((obsi > 31) && (propi != 9))
	{
		//deltaT = 3.; //<-- Working value
		deltaT = 0.;
		errstr << "This is a mosaic pointing observation. "
				"The deltaT value to search for the MOS clock reset time will be set to " << deltaT << " sec "
				<< Sas::warning("mosaicresetdeltaT");
	}

	if (_resetTimes.empty()) {
		const Exposure *exp = parent()->odf()->currentExposure();
		assert(exp);
		MainPeriodicHKParamFinder pf(&oalServer);

		pf.seek(exp->actualStartTime(),
				deltaT,					//	look 90s into the exposure start
										//	to avoid problems with late resets
				exp->actualDuration());	//	... but not beyond end of exposure

		_resetTimes = pf.resetTimes();
		_wrapValue = pf.wrapValue();
	}
} // setStatics()

/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 */
template<>
ObtVector InstrFrameCounterVector<EMOScam>::toObt() const
{
	setStatics();	//	make sure the EPIC reset time/wrap value has been read
					//	from HK

	//
	//	here we need to look in the EMOS Auxiliary File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::AUXILIARY);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdccdNode = auxTable->column("CCDNODE")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data();
	int8 *ccdId = cdccdId->int8Data(),
		 *ccdNode = cdccdNode->int8Data();
	int16 *ftcoarse = cdftcoarse->int16Data(),
		  *ftfine = cdftfine->int16Data();

	int8 ccd = parent()->ccd(),
		 node = parent()->node();

	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();

	double timeRes = XmmPayLoad::emos1.timeResolution();
	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i) {
			if (ccdId[i] == ccd && ccdNode[i] == node &&
				frame[i] == (*this)[n]) {
				//	found it
				unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);
				//
				//	check for wrap-arounds
				//

				int delta = coarseTime - lastCoarseTime;
				if (delta < 0 && -delta > INT(_wrapValue/2) &&
					i != _lastWrapAt) {
					++_nWraps;
					_lastWrapAt = i;
					if (_nWraps >= _resetTimes.size()) {
						errstr<<"There appear to be at least "<<_nWraps+1
							  <<" wrap-arounds in the EMOS frame times but"
							  <<" read only "<<_resetTimes.size()
							  <<" from the HK data - will continue with "
								"nominal wrap-around time increment"
							  <<Sas::warning("TooFewResetsInHk");
						_resetTimes.push_back(_resetTimes.back()+_wrapValue);
					}
				}

				lastCoarseTime = coarseTime;
				obtVector.push_back(
						   DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i]))
									* timeRes + _resetTimes[_nWraps]);
							//	FTCOARSE is 15 bit - range 0..32767
				indx = i;
				break;
			}
		}

		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			errstr<<"Could not find frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD/node: "<<UNT(ccd)<<"/"<<UNT(node)
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to "<<INVALID_OBT
				  <<Sas::warning("MissingOrUnusableFrameNumber");
			obtVector.push_back(INVALID_OBT);
			indx = 0;
		}
	}

	delete cdframe;
	delete cdccdId;
	delete cdccdNode;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	return obtVector;
} // InstrFrameCounterVector<EMOScam>::toObt()



//
//	---( EPN )-------------------------------------------------------------
//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 *  Also
 *  undefined reference to `InstrFrameCounterVector<EPNcam>::_wrapValue'
 *  undefined reference to `InstrFrameCounterVector<EPNcam>::_resetTimes'
 *  => = 0 added and tmp
 */
template<> std::vector<double> InstrFrameCounterVector<EPNcam>::_resetTimes = tmp;
template<> unsigned InstrFrameCounterVector<EPNcam>::_wrapValue = 0;
template<>unsigned short InstrFrameCounterVector<EPNcam>::_chopperValue = 0;

/* AI 2009-07-08
 * ERM only. Not used for any other instrument
 */
template<> std::string InstrFrameCounterVector<EPNcam>::_auxName = "";
template<>
void InstrFrameCounterVector<EPNcam>::setAux(std::string name) const
{
	_auxName = name;
}


template<>
void InstrFrameCounterVector<EPNcam>::setStatics() const
{


	// Check if this is a mosaic poiting observation. If this is the case, we have to set the deltaT value to search
	// for the MOS reset clock values to 1 sec. Otherwise, we can have problems set the time tags correctly for the
	// stable pointing observations
	string prop = parent()->odf()->proposalId();
	int propi = atoi(prop.substr(0,1).c_str());
	string obsid = parent()->odf()->observationId() ;

	int obsi;
	obsi  = atoi(obsid.substr(2).c_str());
	double deltaT = 90.;
	if ((obsi > 31) && (propi != 9))
	{
		//deltaT = 6.; // Working value
		deltaT = 0.;
		errstr << "This is a mosaic pointing observation. "
				"The deltaT value to search for the MOS clock reset time will be set to " << deltaT << " sec "
				<< Sas::warning("mosaicresetdeltaT");
	}



	if (_resetTimes.empty()) {
		const Exposure *exp = parent()->odf()->currentExposure();
		assert(exp);
		MainPeriodicHKParamFinder pf(&oalServer);

		pf.seek(exp->actualStartTime(),
				deltaT,					//	look 90s into the exposure start
										//	to avoid problems with late resets
				exp->actualDuration());	//	... but not beyond end of exposure

		_resetTimes = pf.resetTimes();
		_wrapValue = pf.wrapValue();

		unsigned short quad = EPNcam::toQuadrantId(parent()->ccd());
		AdditionalPeriodicHKParamFinder apf(&oalServer);
	
		apf.seek(exp->actualStartTime(), deltaT, exp->actualDuration());
		_chopperValue = apf.chopperValue(quad);
	}
} // InstrFrameCounterVector<EPNcam>::setStatics()

//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 */

template<>
ObtVector InstrFrameCounterVector<EPNcam>::toObt() const
{
	setStatics();	//	make sure the EPIC reset time has been
					//	read from HK

	int8	ccd = EPNcam::toCcdId(parent()->ccd()),
			quad = EPNcam::toQuadrantId(parent()->ccd());
	EpnAuxDataServer aux(parent(), quad, ccd, _chopperValue, _wrapValue,
						 _resetTimes);

	ObtVector obt(this);
	unsigned N = size();
	obt.reserve(N);

	for (unsigned n=0; n<N; ++n) {
		//
		//	locate AUX data index for this frame
		//
		int i = aux.locateFrame((*this)[n]);
		if (i>=0 && aux.status()[i] & EpnAuxDataServer::NORMAL) {
			obt.push_back(aux.getObt(i));
			SAS_DEBUG(cout<<n<<' '<<setprecision(16)<<obt.back()<<endl;)

		} else {
			//
			//	we have not found the time of a frame counter or the
			//	frame is not usable
			//
			errstr<<"Could not find or use frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD/Quadrant: "<<UNT(ccd)<<"/"<<UNT(quad)
				  <<" ["<<parent()->ccd()<<"]"
				  <<" in auxiliary file `"<<aux.tableName()<<"'"
				  <<" - will set OBT to "<<INVALID_OBT
				  <<Sas::message(Msg::App, Msg::Verbose);
			obt.push_back(INVALID_OBT);
		}
	}

	SAS_DEBUG(cout<<"OBT tags and deltas\n";
			  for(unsigned n=0; n<N; ++n)
				cout<<n<<' '<<setprecision(12)<<obt[n]<<' '
					<<(n>0 ? obt[n]-obt[n-1] : 0.)<<endl;)

	_resetTimes.clear();

	return obt;
} // InstrFrameCounterVector<EPNcam>::toObt()


//
//	---( RGS )--------------------------------------------------------------
//
//
//	other than EMOS and EPN, RGS and OM use the COBT values directly
//	to time stamp frame readout times - no need to read a reset value
//
//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 */

template<>
void InstrFrameCounterVector<RGScam>::setStatics() const
{
	_resetTimes.push_back(0.);
}

template<> unsigned InstrFrameCounterVector<RGScam>::_nWraps = 0;
template<> unsigned InstrFrameCounterVector<RGScam>::_lastWrapAt = 0;
/* AI gcc4 porting
 * undefined reference to `InstrFrameCounterVector<RGScam>::_resetTimes
 * => = 0 added and tmp
 */
template<> std::vector<double> InstrFrameCounterVector<RGScam>::_resetTimes = tmp;
template<> unsigned short InstrFrameCounterVector<RGScam>::_chopperValue = 0;

/* AI 2009-07-08
 * ERM only. Not used for any other instrument
 */
template<> std::string InstrFrameCounterVector<RGScam>::_auxName = "";
template<>
void InstrFrameCounterVector<RGScam>::setAux(std::string name) const
{
	_auxName = name;
}


template<>
ObtVector InstrFrameCounterVector<RGScam>::toObt() const
{
	setStatics();	//	set reset time to 0.

	//
	//	here we need to look in the RGS Auxiliary File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::AUXILIARY);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdftcoarse = auxTable->column("EOSCOARS")->data(),
			   *cdftfine = auxTable->column("EOSFINE")->data();
	int32 *frame = cdframe->int32Data();
	int8 *ccdId = cdccdId->int8Data();
	int32 *ftcoarse = cdftcoarse->int32Data();
	uint16 *ftfine = cdftfine->uint16Data();

	int8 ccd = parent()->ccd();

	ObtVector obtVector(this);
	unsigned N = size();
	obtVector.reserve(N);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();
	double timeRes = XmmPayLoad::rgs1.timeResolution();
	for (unsigned n=0; n<N; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i)
			if (ccdId[i] == ccd && frame[i] == (*this)[n]) {
				//	found it

				unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);
				//
				//	check for wrap-arounds
				//
				if (coarseTime < lastCoarseTime && _lastWrapAt != i
						&& ( lastCoarseTime -  coarseTime >  320000.)) {
					++_nWraps;
					_lastWrapAt = i;
				}

				lastCoarseTime = coarseTime;
				obtVector.push_back(
						   DBL(_nWraps) * 65536. * 65536. +
						   DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i]))* timeRes);
				indx = i;
				break;
			}
		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			//
			errstr<<"Could not find frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD: "<<UNT(ccd)
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to "<<INVALID_OBT
				  <<Sas::warning("MissingFrameNumber");
			obtVector.push_back(INVALID_OBT);
			indx = 0;
		}

	}

	//	tidy up
	delete cdframe;
	delete cdccdId;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	//
	//	check that OBT values lie before end of exposure; if not we had
	//	a CDMU crash (see comment above for PeriodicHKParamFinder::resetTime())
	//	and need to correct all values
	//
	if (obtVector.size() > 0) {
		STime tstart(parent()->odf()->currentExposure()->actualStartTime()),
			  tend(parent()->odf()->currentExposure()->actualEndTime());
		const TimeCorrelator *timeCorrelator = parent()->timeCorrelator();
		STime tfirst(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(obtVector[0]));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-tfirst,
			   corrT = 0.;
		if (deltaT<0.) {
			errstr<<"The On-Board-Time values corresponding to the RGS "
					"frame counters lie after the end of the current exposure "
					"at "<<tend<<". The most likely cause of this is a "
					"CDMU crash that occured during the exposure. Will "
					"adjust all times such that the time of the first frame "
					"coincides with the start of the the exposure at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags shall "
					"be afflicted with the same absolute error."
				  <<Sas::warning("PossiblyInaccurateOBTValues");
			corrT = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime) -
					obtVector[0];
		}
		unsigned N = obtVector.size();
		for (unsigned n=0; n<N; obtVector[n++] += corrT) ;
	}

	return obtVector;
} // InstrFrameCounterVector<RGScam>::toObt()



//
//	---( OM )--------------------------------------------------------------
//
//
//	other than EMOS and EPN, RGS and OM use the COBT values directly
//	to time stamp frame readout times - no need to read a reset value
//
//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<> statement added
 */

template<>
void InstrFrameCounterVector<OMcam>::setStatics() const
{
	_resetTimes.push_back(0.);
}

template<> unsigned InstrFrameCounterVector<OMcam>::_nWraps = 0;
template<> unsigned InstrFrameCounterVector<OMcam>::_lastWrapAt = 0;
/* AI gcc4 porting
 *  undefined reference to `InstrFrameCounterVector<OMcam>::_resetTimes'
 *  => = 0 added and tmp
 */
template<> std::vector<double> InstrFrameCounterVector<OMcam>::_resetTimes = tmp;
template<> unsigned short InstrFrameCounterVector<OMcam>::_chopperValue = 0;

/* AI 2009-07-08
 * ERM only. Not used for any other instrument
 */
template<> std::string InstrFrameCounterVector<OMcam>::_auxName = "";
template<>
void InstrFrameCounterVector<OMcam>::setAux(std::string name) const
{
	_auxName = name;
}


template<>
ObtVector InstrFrameCounterVector<OMcam>::toObt() const
{
	setStatics();	//	set reset time to 0.

	Xmm::DataMode currentMode = parent()->dataMode();
	if (currentMode != Xmm::DATA_FAST)
		errstr<<"Converting "<<Xmm::toString(Xmm::OM)<<" frame counters "
				"makes only sense in "<<Xmm::toString(Xmm::FAST)<<" mode"
				"current mode is: `"<<Xmm::toString(currentMode)<<"'\n"
			  <<Sas::error("IllegalMode");

	//
	//	here we need to look in the OM Fast Mode Event File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::DATA_FAST);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data(),
		  *ftcoarse = cdftcoarse->int32Data();
	int16 *ftfine = cdftfine->int16Data();

	ObtVector obtVector(this);
	unsigned N = size();
	obtVector.reserve(N);
	int16 lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();
	double timeRes = XmmPayLoad::om.timeResolution();
	for (unsigned n=0; n<N; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i)
			if (frame[i] == (*this)[n]) {
				//	found it
				int16 coarseTime = ftcoarse[i];
				//
				//	check for wrap-arounds
				//
				lastCoarseTime = coarseTime;
				if (coarseTime < lastCoarseTime && _lastWrapAt != i) {
					++_nWraps;
					_lastWrapAt = i;
				}
				obtVector.push_back(
						   DBL(_nWraps) * 16384. +
						   DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i])) * timeRes);
				indx = i;
				break;
			}

		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			errstr<<"Could not find time for frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to "<<INVALID_OBT
				  <<Sas::warning("MissingFrameNumber");
			obtVector.push_back(INVALID_OBT);
			indx = 0;
		}

	}

	delete cdframe;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	//
	//	check that OBT values lie before end of exposure; if not we had
	//	a CDMU crash (see comment above for PeriodicHKParamFinder::resetTime())
	//	and need to correct all values
	//
	if (obtVector.size() > 0) {
		STime tstart(parent()->odf()->currentExposure()->actualStartTime()),
			  tend(parent()->odf()->currentExposure()->actualEndTime());
		const TimeCorrelator *timeCorrelator = parent()->timeCorrelator();
		STime tfirst(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(obtVector[0]));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-tfirst,
			   corrT = 0.;
		if (deltaT<0.) {
			errstr<<"The On-Board-Time values corresponding to the OM "
					"time tags lie after the end of the current exposure "
					"at "<<tend<<". The most likely cause of this is a "
					"CDMU crash that occured during the exposure. Will "
					"adjust all times such that the time of the first event "
					"coincides with the start of the the exposure at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags shall "
					"be afflicted with the same absolute error."
				  <<Sas::warning("PossiblyInaccurateOBTValues");
			corrT = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime) -
					obtVector[0];
		}
		unsigned N = obtVector.size();
		for (unsigned n=0; n<N; obtVector[n++] += corrT) ;
	}

	return obtVector;
} // InstrFrameCounterVector<OMcam>::toObt()



//
//	---( ERM )-------------------------------------------------------------
//
template<> unsigned InstrFrameCounterVector<ERMcam>::_nWraps = 0;
template<> unsigned InstrFrameCounterVector<ERMcam>::_lastWrapAt = 0;
template<> std::vector<double> InstrFrameCounterVector<ERMcam>::_resetTimes=tmp;
template<> unsigned InstrFrameCounterVector<ERMcam>::_wrapValue = 0;
template<> unsigned short InstrFrameCounterVector<ERMcam>::_chopperValue = 0;

template<>
void InstrFrameCounterVector<ERMcam>::setStatics() const
{
	_resetTimes.push_back(0.);

} // setStatics()

/* AI 2009-07-08
 * ERM only. Not used for any other instrument
 */
template<> std::string InstrFrameCounterVector<ERMcam>::_auxName="";
template<>
void InstrFrameCounterVector<ERMcam>::setAux(std::string name) const
{
	_auxName = name;
}

template<>
ObtVector InstrFrameCounterVector<ERMcam>::toObt() const
{
	setStatics();	//	make sure the ERM reset time/wrap value has been read
	//	from HK

	//
	//	here we need to look in the ERM Auxiliary File
	//

	DataSet *auxSet = dataSetServer->open(_auxName, DataSet::Read,
										  DataSetServer::Low);
	errstr <<"ERM Auxiliary file: "<< auxSet->name()<<endl;
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data();
	int32 *ftcoarse = cdftcoarse->int32Data();
	uint16 *ftfine = cdftfine->uint16Data();


	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
	  nRecs = auxTable->rows();

	double timeRes = XmmPayLoad::erm.timeResolution();

	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i) {
		  if ( frame[i] == (*this)[n]) {
		    //	found it

		    unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);

		    //
		    //	check for wrap-arounds
		    //
		    int delta = coarseTime - lastCoarseTime;
		    if (delta < 0 && -delta > INT(_wrapValue/2) && i != _lastWrapAt) {
		      ++_nWraps;
		      _lastWrapAt = i;

		      if (_nWraps >= _resetTimes.size()) {
			errstr<<"There appear to be at least "<<_nWraps+1
			      <<" wrap-arounds in the ERM frame times but"
			      <<" read only "<<_resetTimes.size()
			      <<" from the HK data - will continue with "
			  "nominal wrap-around time increment"
			      <<Sas::warning("TooFewResetsInHk");

			_resetTimes.push_back(_resetTimes.back()+_wrapValue);
		      }
		    }

		    lastCoarseTime = coarseTime;

		    obtVector.push_back(
					DBL(coarseTime) +
					DBL(OalDalUtils::unsigned32(ftfine[i]))
					* timeRes + _resetTimes[_nWraps]);
		    //	FTCOARSE is 15 bit - range 0..32767
		    indx = i;
		    break;
		  }
		}

		if (i >= nRecs) {
		  //
		  //	we have not found the time of a frame counter
		  //	let's bail out
		  errstr<<"Could not find frame number "
			<<OalDalUtils::unsigned32((*this)[n])
			<<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
			<<" - will set OBT to "<<INVALID_OBT
			<<Sas::warning("MissingOrUnusableFrameNumber");
		  obtVector.push_back(INVALID_OBT);
		  indx = 0;
		}
	}

	delete cdframe;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	return obtVector;
} // InstrFrameCounterVector<EMOScam>::toObt()







//
//	---( ObtVector )--------------------------------------------------------
//
std::vector<double> ObtVector::toMET() const
{
	unsigned nmax = size();
	std::vector<double> tags(nmax);

	const TimeCorrelator *timeCorrelator =parent()->parent()->timeCorrelator();

	for (unsigned n=0; n<nmax; ++n) {
		if ((*this)[n] == INVALID_OBT) {
			errstr<<"OBT vector element #"<<n<<" is invalid ("
				  <<(*this)[n]<<")"
				  <<Sas::warning("InvalidObtValue");
			NaN.set(tags[n]);
			continue;
		}
		tags[n] = timeCorrelator->toMissionEllapsedTime((*this)[n]);
	}

	return tags;
} // ObtVector::toMET()


double InstrumentalTimer::combine(int ftcoarse, int ftfine) const
{
	return DBL(ftcoarse) + _res*DBL(ftfine);
} // InstrumentalTimer::combine()
